#| 

This file contains an example of using stobj to represent an array and
defining a sort function on that array using the stobj-provided
manipulation functions.

Matt Wilding
March 1999

|#  


(in-package "ACL2")

(include-book "/accts/dagreve/ACL/acl2-2.4/acl2-sources/books/arithmetic/top")
(include-book "/accts/dagreve/ACL/acl2-2.4/acl2-sources/books/data-structures/list-defthms")
(include-book "/accts/dagreve/ACL/acl2-2.4/acl2-sources/books/meta/meta")
(include-book "/accts/dagreve/ACL/acl2-2.4/acl2-sources/books/ihs/logops-lemmas")

(disable-theory (theory 'logops-functions))

; A sort implementation using stobj.  First, some needed background lemmas.

(defthm equal-len-n
  (implies (syntaxp (quotep n))
           (equal (equal (len x) n)
                  (if (integerp n)
                      (if (< n 0)
                          nil
                        (if (equal n 0)
                            (atom x)
                          (and (consp x)
                               (equal (len (cdr x)) (- n 1)))))
                    nil))))


;; We try to backchain first to avoid casesplit.
(defthm nth-set-nth
   (equal
    (nth n1 (set-nth n2 v l))
    (if (equal (nfix n1) (nfix n2))
	v
      (nth n1 l)))
  :hints (("goal" :in-theory (enable set-nth nth))))

(defthm nth-set-nth-1
  (implies
    (not (equal (nfix n1) (nfix n2)))
   (equal
    (nth n1 (set-nth n2 v l))
    (nth n1 l))))

(defthm nth-set-nth-2
  (implies
    (equal (nfix n1) (nfix n2))
   (equal
    (nth n1 (set-nth n2 v l))
    v)))

(defthm set-nth-nth
  (implies
   (< (nfix i) (len l))
   (equal (set-nth i (nth i l) l) l)))

(defthm len-set-nth 
  (implies
   (< (nfix i) (len l))
   (equal (len (set-nth i v l)) 
	  (len l))))



;; Now we define a single-threaded object with an array

(defstobj arr
  (vals :type (array (signed-byte 32) (1048576)) :initially 0))

(defthm valsp1-len
  (implies
   (valsp1 vals len)
   (equal (len vals) len)))

(defthm valsp1-set-nth
  (implies
   (and
    (valsp1 vals len)
    (signed-byte-p 32 v)
    (< (nfix i) len))
   (valsp1 (set-nth i v vals) len))
  :hints (("goal" :in-theory (enable signed-byte-p))))
   
(defthm signed-byte-p-nth-valsp1
  (implies
   (valsp1 vals len)
   (iff (signed-byte-p 32 (nth i vals)) (< (nfix i) (len vals))))
  :hints (("goal" :in-theory (enable signed-byte-p))))

(defthm signed-byte-p-nth-valsp1-opened
  (implies
   (valsp1 vals len)
   (and
    (iff (integerp (nth i vals)) (< (nfix i) (len vals)))
    (<= (nth i vals) 2147483647)
    (<= -2147483648 (nth i vals))
    (iff (rationalp (nth i vals)) (< (nfix i) (len vals))))))

; Find the minimum of a subarray of the s.t. array
(defun minarr (arr l h)
  (declare (xargs :stobjs (arr)
		  :measure (max (- (ifix h) (ifix l)) 0)
		  :guard (and (arrp arr) (integerp l) (<= 0 l) (integerp h) (< h (expt 2 20)))
		  :verify-guards nil))
  (if (and (integerp l) (integerp h) (< l h))
      (let ((v (valsi l arr)))
	(let ((minloc (minarr arr (1+ l) h)))
	  (if (<= v (valsi minloc arr)) l minloc)))
    l))

(defthm nth-minarr-properties
  (implies
   (and (integerp l) (integerp h) (valsp1 (car arr) len) (<= 0 l) (< 0 len) (<= l h) (< h (len (car arr))))
   (and
    (rationalp (nth (minarr arr l h) (car arr)))
    (< (minarr arr l h) len)
    (<= l (minarr arr l h))
    (<= 0 (minarr arr l h)))))

(defthm nth-minarr-properties2
  (implies
   (and (integerp l) (integerp h) (valsp1 vals len) (<= 0 l) (< 0 len) (<= l h) (< h (len vals)))
   (and
    (rationalp (nth (minarr (list vals) l h) vals))
    (< (minarr (list vals) l h) len)
    (<= l (minarr (list vals) l h))
    (<= 0 (minarr (list vals) l h)))))

(defthm nth-car-minarr
  (implies
   (integerp l)
   (<= l (minarr arr l h))))

(defthm equal-minarr-base
  (implies
   (and (< x l) (integerp l))
   (and
    (not (equal x (minarr arr l h)))
    (<= x (minarr arr l h)))))

(defthm integerp-minarr
  (implies
   (integerp l)
   (integerp (minarr arr l h))))

(verify-guards minarr)

; Swap two elements of the array
(defun swaparr (arr i1 i2)
  (declare (xargs :stobjs (arr)
		  :guard (and (arrp arr) (unsigned-byte-p 20 i1) (unsigned-byte-p 20 i2))))
  (let ((v1 (valsi i1 arr)) (v2 (valsi i2 arr)))
    (let ((arr (setvalsi i1 v2 arr)))
      (setvalsi i2 v1 arr))))

; Sort the array
(defun sortarr (l h arr)
  (declare (xargs :stobjs (arr)
		  :measure (max (- (ifix h) (ifix l)) 0)
		  :guard (and (arrp arr) (integerp l) (<= 0 l) (integerp h) (< h (expt 2 20)))
		  :guard-hints (("goal" :in-theory (enable unsigned-byte-p)))))
  (if (not (and (integerp l) (integerp h) (< l h)))
      arr
    (let ((min (minarr arr l h)))
      (let ((arr (swaparr arr l min)))
	(sortarr (1+ l) h arr)))))

; We won't bother to guard-check definitions that relate to our spec.

; Add elements to the array
(defun generatearr (init vallist arr)
  (declare (xargs :stobjs (arr)))
  (if (not (consp vallist))
      arr
    (let ((arr (setvalsi init (car vallist) arr)))
      (generatearr (1+ init) (cdr vallist) arr))))

; Construct a list of values of an array
(defun printarr (l h arr)
  (declare (xargs :stobjs (arr)
		  :measure (max (1+ (- (ifix h) (ifix l))) 0)))
  (if (and (integerp l) (integerp h) (<= l h))
      (cons (valsi l arr) (printarr (1+ l) h arr))
    nil))

; Determine whether a subsequence of an array is ordered
(defun orderedparr (l h arr)
  (declare 
   (xargs :stobjs (arr)
	  :measure (max (- (ifix h) (ifix l)) 0)))
  (if (and (integerp l) (integerp h) (< l h))
      (and (<= (valsi l arr) (valsi (1+ l) arr))
	   (orderedparr (1+ l) h arr))
    t))


;; Some lemmas about our functions

(defthm consp-printarr
  (equal 
   (consp (printarr l h arr))
   (and (integerp l) (integerp h) (<= l h))))

(defthm car-printarr
  (equal
   (car (printarr l h arr))
   (if (and (integerp l) (integerp h) (<= l h))
       (valsi l arr)
     nil))))

(defthm printarr-base1
  (implies
   (and (consp arr) (integerp x))
   (equal (printarr x x arr) (list (valsi x arr))))
  :hints (("goal" :expand (printarr x x arr))))

(defthm printarr-base2
  (implies
   (< h l)
   (equal (printarr l h arr) nil)))
  

(defthm nth-car-sortarr-untouched
  (implies
   (and (integerp x) (integerp l) (< x l) (<= 0 x))
   (equal (nth x (car (sortarr l h arr)))
	  (nth x (car arr))))
  :hints (("goal" :in-theory (disable minarr))))

(defthm nth-car-sortarr-first
  (implies
   (and (integerp l) (integerp h) (<= l h) (<= 0 l))
   (equal (nth l (car (sortarr l h arr)))
	  (nth (minarr arr l h) (car arr))))
  :hints (("goal" :expand (sortarr l h arr))))

(defthm minarr-bound-lower
  (<= l (minarr arr l h)))

(defthm minarr-bound-upper
  (implies
   (<= l h)
   (<= (minarr arr l h) h)))

(defthm <-member-minarr
  (implies
   (and (integerp l) (integerp i) (integerp h) (<= l i) (<= i h))
   (<= (nth (minarr arr l h) (car arr))
       (nth i (car arr)))))

(defthm minarr-bound
  (implies
   (and (<= l h) (< h (len (car arr))) (integerp l) (integerp h))
   (< (minarr arr l h) (len (car arr)))))

(defthm minarr-bound2
  (implies
   (and (valsp1 (car arr) 1048576) (<= l h) (< h (len (car arr))) (integerp l) (integerp h))
   (< (minarr arr l h) 1048576)))

(defthm orderedpparr-sortarr-helper
  (implies
   (and (arrp arr) 
	(integerp l) (integerp h) (<= 0 l) (<= l h) 
	(< h (len (car arr)))  (< l (len (car arr)))) ; redundant
   (orderedparr l h (sortarr l h arr))))


;; Sorting the array puts it in order

(defthm orderedparr-sortarr
  (implies
   (and (arrp arr) 
	(integerp l) (integerp h) (<= 0 l) (<= l h) 
	(< h (len (car arr))))
   (orderedparr l h (sortarr l h arr))))

;; We choose to recast orderedp as a predicate on lists

(defun orderedp (l)
  (if (consp l)
      (if (consp (cdr l))
	  (and
	   (<= (car l) (cadr l))
	   (orderedp (cdr l)))
	t)
    t))

(defthm orderedparr-as-orderedp
  (equal
   (orderedparr l h arr)
   (orderedp (printarr l h arr))))

(defun permlistp (l1 l2)
  (and
   (subbagp l1 l2)
   (subbagp l2 l1)))

(defun occurs (e l)
  (if (consp l)
      (if (equal e (car l))
	  (1+ (occurs e (cdr l)))
	(occurs e (cdr l)))
    0))

(defthm memb-as-occurs
  (equal
   (memb e l)
   (< 0 (occurs e l))))

(defthm occurs-del
  (equal
   (occurs e (del e l))
   (if (memb e l) (1- (occurs e l)) 0)))

(defthm occurs-del2
  (implies
   (not (equal e1 e2))
   (equal (occurs e1 (del e2 l)) (occurs e1 l))))

;; Matt Kaufmann's witness trick.  This idea appears in Nqthm-1992 examples.
(defun badguy (x y)
  (if (consp x)
      (if (memb (car x) y)
          (badguy (cdr x) (del (car x) y))
	(car x))
    0))

(defthm subbagp-wit-lemma
  (equal (subbagp x y)
         (<= (occurs (badguy x y) x)
	     (occurs (badguy x y) y))))

(defthm lessp-1+-hack
  (implies
   (and
    (integerp x)
    (integerp y)
    (< x y))
   (not (< y (1+ x)))))

(defthm swaparr-base
  (implies
   (< (nfix l) (len (car arr)))
   (equal (swaparr arr l l) arr)))

(defthm len-car-swaparr
  (implies
   (and (integerp i1) (integerp i2) (<= 0 i1) (<= 0 i2) (< i1 (len (car arr))) (< i2 (len (car arr))))
   (equal (len (car (swaparr arr i1 i2))) (len (car arr)))))

(defthm printarr-set-nth-outer
  (implies
   (and (< i l) (integerp i) (<= 0 i))
   (equal (printarr l h (list (set-nth i v vals)))
	  (printarr l h (list vals)))))

(defthm occurs-printarr-set-nth
  (implies
   (and (integerp l) (integerp h) (integerp i) (<= 0 l) (<= l i) (<= i h))
  (equal
   (occurs e (printarr l h (list (set-nth i v vals))))
   (if (equal e v)
       (if (equal (nth i vals) e)
	   (occurs e (printarr l h (list vals)))
	 (1+ (occurs e (printarr l h (list vals)))))
     (if (equal (nth i vals) e)
	 (1- (occurs e (printarr l h (list vals))))
       (occurs e (printarr l h (list vals))))))))

(defthm occurs-printarr-swaparr
  (implies
   (and (arrp arr) (integerp l1) (integerp l2) (integerp h1) (integerp h2) 
	(<= l1 l2) (<= l2 h2) (<= h2 h1) (<= 0 l1))
   (equal
    (occurs e (printarr l1 h1 (swaparr arr l2 h2)))
    (occurs e (printarr l1 h1 arr)))))

(defthm true-listp-swaparr
  (implies
   (true-listp arr)
   (true-listp (swaparr arr i1 i2))))

(defthm valsp1-car-swaparr
  (implies
   (and
    (valsp1 (car arr) len)
    (< (nfix i1) len)
    (< (nfix i2) len)
    (integerp i1) (integerp i2)
    (<= 0 i1) (<= 0 i2)
    (consp arr))
   (valsp1 (car (swaparr arr i1 i2)) len))
  :hints (("goal" :in-theory (enable signed-byte-p))))

(defthm cdr-swaparr
  (equal (cdr (swaparr arr l h)) (cdr arr)))

(defthm minarr-pos
  (implies
   (<= 0 l)
   (<= 0 (minarr arr l h))))

(defthm minarr-upper-hack
  (implies
   (and
    (<= h h2)
    (<= l h))
   (<= (minarr arr l h) h2)))

(defthm occurs-printarr-sortarr
  (implies
   (and (arrp arr) (integerp l1) (integerp l2) (integerp h1) (integerp h2)
	(<= 0 l2) (<= 0 l1) (<= l1 l2) (<= h2 h1) (<= l2 h2) (< l2 (len (car arr))) 
	(< h1 (len (car arr))) (< h2 (len (car arr))) )  ; redundant hyps help proof
  (equal
   (occurs e (printarr l1 h1 (sortarr l2 h2 arr)))
   (occurs e (printarr l1 h1 arr))))
  :hints (("goal" :induct (sortarr l2 h2 arr) :in-theory (disable swaparr))))

;; **** Part of what it means for sort to work ****

(defthm permlistp-sortarr
  (implies
   (and (arrp arr) 
	(integerp l) (integerp h) (<= 0 l) (<= l h) 
	(< h (len (car arr))))
   (permlistp (printarr l h arr)
	      (printarr l h (sortarr l h arr)))))

(defthm ordered-sortarr
  (implies
   (and (arrp arr) 
	(integerp l) (integerp h) (<= 0 l) (<= l h) 
	(< h (len (car arr))))
   (orderedp (printarr l h (sortarr l h arr))))
  :hints (("goal" :use (:instance orderedparr-sortarr))))

