#|

This file contains an interpreter for the "tiny" machine, an
experimental machine loosely based on Boyer's and Moore's small
machine.  It is the latest in a series of experiments to build a toy
interpreter that can be analyzed and run fast.  

Some of the earlier methods we've employed in this experiment, as well
as our motivation for focussing on ACL2 execution speed, a description
of the toy processor, and an indication of how we apply these
techniques on real processors, are documented in

  M. Wilding, D. Greve, D. Hardin, "Efficient Simulation of Formal
  Processor Models" 1998.  (www.pobox.com/~hokie/docs/efm.ps.)

There are two innovations in this version.  First, we make use of
stobj, the ACL2 single-threaded object implemented by J Moore in ACL2
2.4, to ensure the soundness of destructive data structure access
needed for fast execution.  This eliminates the two modes from ealier
versions of the model, and relies on ACL2 to guarantee the soundness
of the destructive update optimization.  Second, this model was
updated to use Bishop's IHS definitions.

Dave Greve and Matt Wilding 
March 1999

|#
(in-package "ACL2")

;; Normally we now load some subsidiary books.  But we will assume
;; that they are already loaded to simplify the make process.

;;(include-book "/accts/dagreve/ACL/acl2-2.4/acl2-sources/books/arithmetic/top")
;;(include-book "/accts/dagreve/ACL/acl2-2.4/acl2-sources/books/data-structures/list-defthms")
;;(include-book "/accts/dagreve/ACL/acl2-2.4/acl2-sources/books/meta/meta")
;;(include-book "/accts/dagreve/ACL/acl2-2.4/acl2-sources/books/ihs/logops-lemmas")

(disable-theory (theory 'logops-functions))
(set-verify-guards-eagerness 2) 

(defstobj st
          (progc :type (unsigned-byte 10) :initially 0)
	  (mem :type (array (signed-byte 32) (1024)) :initially 0)
	  (dtos :type (unsigned-byte 10) :initially 0)
	  (ctos :type (unsigned-byte 10) :initially 0))

(defmacro bv32 (x) `(the (signed-byte 32) ,x))

(defmacro memaddrfix (x) `(the (unsigned-byte 10) ,x))
(defmacro memvalfix (x) `(bv32 ,x))

(defthm integerp-plus
  (implies
   (and (integerp a) (integerp b))
   (integerp (+ a b))))

(defun negatebv32 (x)
  (declare (type (signed-byte 32) x))
  (logext 32 (- x)))

(defun +bv32 (x y)
   (declare (type (signed-byte 32) x y))
   (logext 32 (+ x y)))

(defun -bv32 (x y)
  (declare
   (xargs :guard (and (integerp x) (integerp y))
	  :guard-hints (("goal" :in-theory (set-difference-theories (enable signed-byte-p) '(signed-byte-p-logext))
			 :use ((:instance signed-byte-p-logext (size 32) (size1 32) (i (- y)))
			       (:instance signed-byte-p-logext (size 32) (size1 32) (i (+ X (- Y))))))))
   (type (signed-byte 32) x y))
  (bv32 (+bv32 (bv32 x) (bv32 (negatebv32 (bv32 y))))))

(defthm integerp-means-rationalp
  (implies (integerp x) (rationalp x))
  :rule-classes nil)

(defthm +bv32-representable-as-32-bits
  (and
   (integerp (+bv32 a b))
   (rationalp (+bv32 a b))
   (<= (+bv32 a b) 2147483647)
   (>= (+bv32 a b) -2147483648))
  :hints (("goal" 
	   :in-theory (set-difference-theories (enable signed-byte-p) '(integerp-plus signed-byte-p-logext))
	   :use
	   ((:instance integerp-means-rationalp
		       (x (+bv32 a b)))
	    (:instance signed-byte-p-logext (size 32) (size1 32) (i (+ a b)))
	    (:instance integerp-plus
		       (a (LOGAND A 4294967295))
		       (b (LOGAND B 4294967295)))))))

(in-theory (disable +bv32))

(defmacro addr (address delta)
  `(memaddrfix (logand (bv32 (+bv32 (memvalfix ,address) (memvalfix ,delta))) 1023)))

;; swiped from J's documentation.  syntaxp rules!
(defthm equal-len-n
  (implies (syntaxp (quotep n))
           (equal (equal (len x) n)
                  (if (integerp n)
                      (if (< n 0)
                          nil
                        (if (equal n 0)
                            (atom x)
                          (and (consp x)
                               (equal (len (cdr x)) (- n 1)))))
                    nil))))

(defthm set-nth-memp1 
  (implies
   (and
    (memp1 mem n2)
    (< n n2)
    (integerp n)
    (>= n 0)
    (integerp n2))
   (equal (memp1 (set-nth n val mem) n2)
	  (and
	   (<= (- (expt 2 31)) val)
	   (<= val (1- (expt 2 31)))
	   (integerp val))))
  :hints (("goal" :in-theory (enable set-nth))))

(defthm arb-memory-proof 
  (implies
   (and
    (memp1 mem n)
    (< n2 n)
    (<= 0 n2))
   (and
    (<= (nth n2 mem) 2147483647)
    (<= -2147483648 (nth n2 mem))
    (integerp (nth n2 mem))))
  :rule-classes nil)

(defthm arb-memory
  (implies
   (and
    (memp1 mem n)
    (< n2 n)
    (<= 0 n2))
   (and
    (<= (nth n2 mem) 2147483647)
    (not (< 2147483648 (NTH n2 mem))) ; needed later on.  Geez
    (<= -2147483648 (nth n2 mem))
    (integerp (nth n2 mem))
    (acl2-numberp (nth n2 mem))))
  :hints (("goal" :use arb-memory-proof)))

(defun pushus (val st)
  (declare (type (signed-byte 32) val)
	   (xargs :stobjs (st)
		  :guard (stp st)))
  (let ((dtos (dtos st)))
    (let ((st (setmemi dtos (memvalfix val) st)))
     (setdtos (addr dtos -1) st))))

(defun popus (address st)
  (declare (type (unsigned-byte 10) address)
	   (xargs :stobjs (st)
		  :guard (stp st)))
  (let ((ndtos (addr (dtos st) 1)))
    (let ((st (setmemi address (memi ndtos st) st)))
      (setdtos ndtos st))))

(defun pushcs (val st)
  (declare (type (signed-byte 32) val)
	   (xargs :stobjs (st)
		  :guard (stp st)))
  (let ((ctos (ctos st)))
    (let ((st (setmemi ctos (memvalfix val) st)))
     (setctos (addr ctos -1) st))))

(defun popcs (address st)
  (declare (type (unsigned-byte 10) address)
	   (xargs :stobjs (st)
		  :guard (stp st)))
  (let ((nctos (addr (ctos st) 1)))
    (let ((st (setmemi address (memi nctos st) st)))
      (setdtos nctos st))))


(defun next (st)
  (declare (xargs :stobjs (st)
		  :guard (stp st)
		  :guard-hints (("goal" :in-theory
				 (disable hard-error pushus popus pushcs)))))
  (let ((progc (progc st))
	(dtos (dtos st))
	(ctos (ctos st)))
    (let ((ins (memi progc st)))
      (case ins
	; pop
	(0 (let ((st (setprogc (addr progc 2) st)))
	    (popus (addr (memi (addr progc 1) st) 0) st)))

	; pushs
	(1 (let ((st (setprogc (addr progc 2) st)))
	    (pushus (memi (addr (memi (addr progc 1) st) 0) st) st)))

	; pushsi
	(2 (let ((st (setprogc (addr progc 2) st)))
	    (pushus (memi (addr progc 1) st) st)))

	; add
	(3 (let ((st (setprogc (addr progc 1) st)))
	     (let ((st (setdtos (addr dtos 2) st)))
	       (pushus (bv32 (+bv32 (bv32 (memi (addr dtos 1) st))
			            (bv32 (memi (addr dtos 2) st))))
		       st))))

	; jump
	(4 (setprogc (addr (memi (addr progc 1) st) 0) st))

	; jumpz
	(5 (let ((st (setprogc
		      (if (= (memi (addr dtos 1) st) 0)
			  (addr (memi (addr progc 1) st) 0)
			(addr progc 2))
		      st)))
	     (setdtos (addr dtos 1) st)))

	; call
	(6 (let ((nadd (memi (addr progc 1) st)))
	     (declare (type (signed-byte 32) nadd))
	     (let ((st (setprogc (addr nadd 0) st)))
	      (pushcs (addr progc 2) st))))

	; return
	(7 (let ((nadd (memi (addr ctos 1) st)))
	     (declare (type (signed-byte 32) nadd))
	     (let ((st (setprogc (addr nadd 0) st)))
	      (setctos (addr ctos 1) st))))

	; sub
	(8 (let ((st (setprogc (addr progc 1) st)))
	     (let ((st (setdtos (addr dtos 2) st)))
	       (pushus (bv32 (max 0 (-bv32 (memi (addr dtos 2) st)
					   (memi (addr dtos 1) st))))
		       st))))

	; dup
	(9 (let ((st (setprogc (addr progc 1) st)))
	    (pushus (memi (addr dtos 1) st) st)))

	; halt
	(otherwise st)))))


(defthm stp-next
  (implies
   (stp st)
   (stp (next st))))

(defun tiny (st n)
  (declare (xargs :stobjs (st)
		  :guard (stp st)
		  :guard-hints (("goal" :in-theory (disable next stp))))
	   (type (signed-byte 32) n))
  (if (or (not (integerp (bv32 n))) (<= (bv32 n) (bv32 0)))   ; ifix inefficient
      st
    (let ((st (next st)))
      (tiny st (bv32 (1- (bv32 n)))))))


;;;;;
;; Functions for creating Tiny state

;; Recursive functions involving s.t. objects must have a base case
;; listed first.  This allows ACL2 a simpler approach for determining
;; whether a function returns all the s.t. objects it needs to return,
;; since every base case must return every potentially-modified
;; s.t. object.

(defun load-memory-block (address list st)
  (declare (xargs :stobjs (st)
		  :guard (stp st)
		  :verify-guards nil))
  (if (not (consp list)) st
      (let ((st (setmemi address (car list) st)))
	(load-memory-block (addr address 1) (cdr list) st))))
  
(defun load-memory (assoc st)
  (declare (xargs :stobjs (st)
		  :verify-guards nil))
  (if (not (consp assoc)) st
    (let ((st (load-memory-block (caar assoc) (cdar assoc) st)))
      (load-memory (cdr assoc) st))))


(defun load-tiny (pc memlist st)
  (declare (xargs :stobjs (st)
		  :verify-guards nil))
  (let ((st (setprogc pc st)))
    (let ((st (setdtos 900 st)))
      (let ((st (setctos 1000 st)))
	(load-memory memlist st)))))

(defun encode (op)
  (declare (xargs :verify-guards nil))
  (case op
    ('pop 0)
    ('pushs 1)
    ('pushsi 2)
    ('add 3)
    ('jump 4)
    ('jumpz 5)
    ('call 6)
    ('ret 7)
    ('sub 8)
    ('dup 9)
    ('halt 10)))

(defun assemble (code)
  (declare (xargs :verify-guards nil))
  (if (consp code)
      (cons
       (if (integerp (car code)) (car code) (encode (car code)))
       (assemble (cdr code)))
    nil))

(defconst *mod-caller-prog* (assemble '(pushsi 0 pushsi 1 add pushsi 100 call 20 jump 6)))


(defconst *mod-prog* (assemble '(pop 19   ;20
				 pop 18
				 pushs 18 ;24
				 pushs 19
				 sub
				 dup
				 jumpz 36
				 pop 18
				 jump 24
				 pushs 19 ;36
				 pushs 18
				 sub
				 jumpz 45
                                 jump 49
				 pushsi 0 ;45 
				 pop 18
				 pop 19   ;49
				 pushs 18
				 ret
				 )))

(defun run-example-tiny-state (n st)
  (declare (xargs :verify-guards nil :stobjs (st)))
  (let ((st (load-tiny 4 (list (cons 4 *mod-caller-prog*) (cons 20 *mod-prog*)) st)))
    (tiny st n)))
